# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .field import Field, FieldException
from .mapper import ModelAttributeAdapter
from sqlalchemy.schema import Sequence as SA_Sequence, Column as SA_Column
from sqlalchemy import types, CheckConstraint
from sqlalchemy_utils.types.color import ColorType
from sqlalchemy_utils.types.encrypted.encrypted_type import EncryptedType
from sqlalchemy_utils.types.password import PasswordType, Password as SAU_PWD
from sqlalchemy_utils.types.uuid import UUIDType
from sqlalchemy_utils.types.url import URLType
from sqlalchemy_utils.types.phone_number import PhoneNumberType
from sqlalchemy_utils.types.email import EmailType
from sqlalchemy_utils.types.scalar_coercible import ScalarCoercible
from datetime import datetime, date
from dateutil.parser import parse
from inspect import ismethod
from anyblok.config import Configuration
import time
import pytz
import decimal
from logging import getLogger
from hashlib import md5


pycountry = None
python_pycountry_type = None
try:
    import pycountry
    if not pycountry.countries._is_loaded:
        pycountry.countries._load()

    python_pycountry_type = pycountry.countries.data_class
except ImportError:
    pass


logger = getLogger(__name__)


def wrap_default(registry, namespace, default_val):

    def wrapper():
        Model = registry.get(namespace)
        if hasattr(Model, default_val):
            func = getattr(Model, default_val)
            if ismethod(func):
                if default_val not in Model.loaded_columns:
                    if default_val not in Model.loaded_fields:
                        return func()
                    else:
                        logger.warn("On a Model %r the attribute %r is "
                                    "declared as a default value, a field "
                                    "with the same name exist" % (namespace,
                                                                  default_val))
                else:
                    logger.warn("On a Model %r the attribute %r is declared "
                                "as a default value, a column with the same "
                                "name exist" % (namespace, default_val))
            else:
                logger.warn("On a Model %r the attribute %r is declared as a "
                            "default value, a instance method with the same "
                            "name exist" % (namespace, default_val))

        return default_val

    return wrapper


class ColumnDefaultValue:

    def __init__(self, callable):
        self.callable = callable

    def get_default_callable(self, registry, namespace, fieldname, properties):
        return self.callable(registry, namespace, fieldname, properties)


class NoDefaultValue:
    pass


class Column(Field):
    """ Column class

    This class can't be instanciated
    """

    use_hybrid_property = True
    foreign_key = None
    sqlalchemy_type = None
    type = None

    def __init__(self, *args, **kwargs):
        """ Initialize the column

        :param label: label of this field
        :type label: str
        """
        self.forbid_instance(Column)
        assert self.sqlalchemy_type
        self.sequence = None

        if 'type_' in kwargs:
            del kwargs['type_']

        if 'foreign_key' in kwargs:
            self.foreign_key = ModelAttributeAdapter(kwargs.pop('foreign_key'))

        if 'sequence' in kwargs:
            self.sequence = SA_Sequence(kwargs.pop('sequence'))

        self.db_column_name = None
        if 'db_column_name' in kwargs:
            self.db_column_name = kwargs.pop('db_column_name')

        self.default_val = NoDefaultValue
        if 'default' in kwargs:
            self.default_val = kwargs.pop('default')

        self.encrypt_key = kwargs.pop('encrypt_key', None)
        super(Column, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        res = super(Column, self).autodoc_get_properties()
        res['foreign_key'] = self.foreign_key
        res['DB column name'] = self.db_column_name
        res['default'] = self.default_val
        res['is crypted'] = True if self.encrypt_key else False
        return res

    autodoc_omit_property_values = Field.autodoc_omit_property_values.union((
        ('foreign_key', None),
        ('DB column name', None),
        ('default', None),
        ('is crypted', False),
    ))

    def native_type(cls):
        """ Return the native SqlAlchemy type """
        return cls.sqlalchemy_type

    def format_foreign_key(self, registry, args, kwargs):
        if self.foreign_key:
            args = args + (self.foreign_key.get_fk(registry),)
            kwargs['info'].update({
                'foreign_key': self.foreign_key.get_fk_name(registry),
                'remote_model': self.foreign_key.model_name,
            })

        return args

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Return the instance of the real field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: known properties of the model
        :rtype: sqlalchemy column instance
        """
        self.format_label(fieldname)
        args = self.args
        kwargs = self.kwargs.copy()
        if 'info' not in kwargs:
            kwargs['info'] = {}
        args = self.format_foreign_key(registry, args, kwargs)
        kwargs['info']['label'] = self.label
        if self.sequence:
            args = (self.sequence,) + args

        if self.db_column_name:
            db_column_name = self.db_column_name
        else:
            db_column_name = fieldname

        if self.default_val is not NoDefaultValue:
            if isinstance(self.default_val, str):
                kwargs['default'] = wrap_default(registry, namespace,
                                                 self.default_val)
            elif isinstance(self.default_val, ColumnDefaultValue):
                kwargs['default'] = self.default_val.get_default_callable(
                    registry, namespace, fieldname, properties)
            else:
                kwargs['default'] = self.default_val

        sqlalchemy_type = self.sqlalchemy_type
        if self.encrypt_key:
            encrypt_key = self.format_encrypt_key(registry, namespace)
            sqlalchemy_type = EncryptedType(sqlalchemy_type, encrypt_key)

        return SA_Column(db_column_name, sqlalchemy_type, *args, **kwargs)

    def format_encrypt_key(self, registry, namespace):
        encrypt_key = self.encrypt_key
        if encrypt_key is True:
            encrypt_key = Configuration.get('default_encrypt_key')

        if not encrypt_key:
            raise FieldException("No encrypt_key define in the "
                                 "configuration")

        def wrapper():
            Model = registry.get(namespace)
            if hasattr(Model, encrypt_key):
                func = getattr(Model, encrypt_key)
                if ismethod(func):
                    if encrypt_key not in Model.loaded_columns:
                        if encrypt_key not in Model.loaded_fields:
                            return func()

            return encrypt_key

        return wrapper

    def must_be_declared_as_attr(self):
        """ Return True if the column have a foreign key to a remote column """
        if self.foreign_key is not None:
            return True

        return False


class Integer(Column):
    """ Integer column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Integer


        @Declarations.register(Declarations.Model)
        class Test:

            x = Integer(default=1)

    """

    def __init__(self, *args, **kwargs):
        super(Integer, self).__init__(*args, **kwargs)
        if self.kwargs.get('primary_key') is True:
            if 'autoincrement' not in self.kwargs:
                self.kwargs['autoincrement'] = True

    sqlalchemy_type = types.Integer


class BigInteger(Column):
    """ Big integer column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import BigInteger


        @Declarations.register(Declarations.Model)
        class Test:

            x = BigInteger(default=1)

    """
    sqlalchemy_type = types.BigInteger


class Boolean(Column):
    """ Boolean column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Boolean


        @Declarations.register(Declarations.Model)
        class Test:

            x = Boolean(default=True)

    """
    sqlalchemy_type = types.Boolean


class Float(Column):
    """ Float column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Float


        @Declarations.register(Declarations.Model)
        class Test:

            x = Float(default=1.0)

    """
    sqlalchemy_type = types.Float


class Decimal(Column):
    """ Decimal column

    ::

        from decimal import Decimal as D
        from anyblok.declarations import Declarations
        from anyblok.column import Decimal


        @Declarations.register(Declarations.Model)
        class Test:

            x = Decimal(default=D('1.1'))

    """
    sqlalchemy_type = types.DECIMAL

    def setter_format_value(self, value):
        if value is not None:
            if not isinstance(value, decimal.Decimal):
                value = decimal.Decimal(value)

        return value


class Date(Column):
    """ Date column

    ::

        from datetime import date
        from anyblok.declarations import Declarations
        from anyblok.column import Date


        @Declarations.register(Declarations.Model)
        class Test:

            x = Date(default=date.today())

    """
    sqlalchemy_type = types.Date


def convert_string_to_datetime(value):
    if value is None:
        return None
    elif isinstance(value, datetime):
        return value
    elif isinstance(value, date):
        return datetime.combine(value, datetime.min.time())
    elif isinstance(value, str):
        return parse(value)

    return value


def add_timezone_on_datetime(dt, default_timezone):
    if dt is not None:
        if dt.tzinfo is None:
            dt = default_timezone.localize(dt)

    return dt


class DateTimeType(types.TypeDecorator):

    impl = types.DateTime(timezone=True)

    def __init__(self, field):
        self.default_timezone = field.default_timezone
        self.field = field

    def process_bind_param(self, value, engine):
        value = convert_string_to_datetime(value)
        value = add_timezone_on_datetime(value, self.default_timezone)
        if self.field.encrypt_key:
            return value.isoformat()

        return value

    def process_result_value(self, value, dialect):
        if self.field.encrypt_key:
            return convert_string_to_datetime(value)

        return value

    @property
    def python_type(self):
        return datetime


class DateTime(Column):
    """ DateTime column

    ::

        from datetime import datetime
        from anyblok.declarations import Declarations
        from anyblok.column import DateTime


        @Declarations.register(Declarations.Model)
        class Test:

            x = DateTime(default=datetime.now)

    """

    def __init__(self, *args, **kwargs):
        self.auto_update = kwargs.pop('auto_update', False)
        default_timezone = kwargs.pop(
            'default_timezone',
            Configuration.get('default_timezone')
        )
        if not default_timezone:
            default_timezone = time.tzname[0]

        if isinstance(default_timezone, str):
            default_timezone = pytz.timezone(default_timezone)

        self.default_timezone = default_timezone
        self.sqlalchemy_type = DateTimeType(self)
        super(DateTime, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        value = convert_string_to_datetime(value)
        return add_timezone_on_datetime(value, self.default_timezone)

    def autodoc_get_properties(self):
        res = super(Column, self).autodoc_get_properties()
        res['is auto updated'] = self.auto_update
        return res


class Time(Column):
    """ Time column

    ::

        from datetime import time
        from anyblok.declarations import Declarations
        from anyblok.column import Time


        @Declarations.register(Declarations.Model)
        class Test:

            x = Time(default=time())

    """
    sqlalchemy_type = types.Time


class Interval(Column):
    """ Datetime interval column

    ::

        from datetime import timedelta
        from anyblok.declarations import Declarations
        from anyblok.column import Interval


        @Declarations.register(Declarations.Model)
        class Test:

            x = Interval(default=timedelta(days=5))

    """
    sqlalchemy_type = types.Interval


class StringType(types.TypeDecorator):

    impl = types.String

    def process_bind_param(self, value, engine):
        if value is False:
            value = ''

        return value

    def process_result_value(self, value, dialect):
        return value


class String(Column):
    """ String column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import String


        @Declarations.register(Declarations.Model)
        class Test:

            x = String(default='test')

    """
    def __init__(self, *args, **kwargs):
        self.size = kwargs.pop('size', 64)
        if 'type_' in kwargs:
            del kwargs['type_']

        self.sqlalchemy_type = StringType(self.size)
        super(String, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        res = super(String, self).autodoc_get_properties()
        res['size'] = self.size
        return res


class Password(Column):
    """ String column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Password


        @Declarations.register(Declarations.Model)
        class Test:

            x = Password(crypt_context={'schemes': ['md5_crypt']})

        =========================================

        test = Test.insert()
        test.x = 'mypassword'

        test.x
        ==> Password object with encrypt value, the value can not be read

        test.x == 'mypassword'
        ==> True

    ..warning::

        the column type Password can not be querying::

            Test.query().filter(Test.x == 'mypassword').count()
            ==> 0
    """
    def __init__(self, *args, **kwargs):
        size = kwargs.pop('size', 64)
        crypt_context = kwargs.pop('crypt_context', {})
        self.crypt_context = crypt_context

        if 'type_' in kwargs:
            del kwargs['type_']

        if 'foreign_key' in kwargs:
            raise FieldException("Column Password can not have a foreign key")

        self.sqlalchemy_type = PasswordType(max_length=size, **crypt_context)
        super(Password, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        value = self.sqlalchemy_type.context.encrypt(value).encode('utf8')
        value = SAU_PWD(value, context=self.sqlalchemy_type.context)
        return value

    def autodoc_get_properties(self):
        res = super(Password, self).autodoc_get_properties()
        res['Crypt context'] = self.crypt_context
        return res


class TextType(types.TypeDecorator):

    impl = types.Text

    def process_bind_param(self, value, engine):
        if value is False:
            value = ''

        return value

    def process_result_value(self, value, dialect):
        return value


class Text(Column):
    """ Text column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Text


        @Declarations.register(Declarations.Model)
        class Test:

            x = Text(default='test')

    """
    sqlalchemy_type = TextType


class StrSelection(str):
    """ Class representing the data of one column Selection """
    selections = {}
    registry = None
    namespace = None

    def get_selections(self):
        if isinstance(self.selections, dict):
            return self.selections
        if isinstance(self.selections, str):
            m = self.registry.get(self.namespace)
            return dict(getattr(m, self.selections)())

    def validate(self):
        a = super(StrSelection, self).__str__()
        return a in self.get_selections().keys()

    @property
    def label(self):
        a = super(StrSelection, self).__str__()
        return self.get_selections()[a]


class SelectionType(types.TypeDecorator):
    """ Generic type for Column Selection """

    impl = types.String

    def __init__(self, selections, size, registry=None, namespace=None):
        super(SelectionType, self).__init__(length=size)
        self.size = size
        if isinstance(selections, (dict, str)):
            self.selections = selections
        elif isinstance(selections, (list, tuple)):
            self.selections = dict(selections)
        else:
            raise FieldException(
                "selection wainting 'dict', get %r" % type(selections))

        if isinstance(self.selections, dict):
            for k in self.selections.keys():
                if not isinstance(k, str):
                    raise FieldException('The key must be a str')
                if len(k) > 64:
                    raise Exception(
                        '%r is too long %r, waiting max %s or use size arg' % (
                            k, len(k), size))

        self._StrSelection = type('StrSelection', (StrSelection,),
                                  {'selections': self.selections,
                                   'registry': registry,
                                   'namespace': namespace})

    @property
    def python_type(self):
        return self._StrSelection

    def process_bind_param(self, value, engine):
        if value is not None:
            value = self.python_type(value)

        return value

    def process_result_value(self, value, dialect):
        return value


class Selection(Column):
    """ Selection column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Selection


        @Declarations.register(Declarations.Model)
        class Test:
            STATUS = (
                (u'draft', u'Draft'),
                (u'done', u'Done'),
            )

            x = Selection(selections=STATUS, size=64, default=u'draft')

    """
    def __init__(self, *args, **kwargs):
        self.selections = tuple()
        self.size = 64
        if 'selections' in kwargs:
            self.selections = kwargs.pop('selections')

        if 'size' in kwargs:
            self.size = kwargs.pop('size')

        self.sqlalchemy_type = 'tmp value for assert'

        super(Selection, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        res = super(Selection, self).autodoc_get_properties()
        res['selections'] = self.selections
        res['size'] = self.size
        return res

    def getter_format_value(self, value):
        if value is None:
            return None

        return self.sqlalchemy_type.python_type(value)

    def setter_format_value(self, value):
        if value is not None:
            val = self.sqlalchemy_type.python_type(value)
            if not val.validate():
                raise FieldException('%r is not in the selections (%s)' % (
                    value, ', '.join(val.get_selections())))

        return value

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        self.sqlalchemy_type = SelectionType(
            self.selections, self.size, registry=registry, namespace=namespace)
        return super(Selection, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)

    def update_description(self, registry, model, res):
        super(Selection, self).update_description(registry, model, res)
        sqlalchemy_type = SelectionType(
            self.selections, self.size, registry=registry, namespace=model)
        values = sqlalchemy_type._StrSelection().get_selections()
        res['selections'] = [(k, v) for k, v in values.items()]

    def must_be_duplicate_before_added(self):
        """ Return True because the field selection in a mixin must be copied
        else the selection method can be wrond
        """
        if isinstance(self.selections, str):
            return True
        else:
            return False

    def update_properties(self, registry, namespace, fieldname, properties):
        super(Selection, self).update_properties(registry, namespace,
                                                 fieldname, properties)
        self.fieldname = fieldname
        properties['add_in_table_args'].append(self)

    def update_table_args(self, Model):
        """return check constraint to limit the value"""
        selections = self.sqlalchemy_type.selections
        if isinstance(selections, dict):
            enum = selections.keys()
        else:
            enum = getattr(Model, selections)()
            if isinstance(enum, (list, tuple)):
                enum = dict(enum)

            enum = enum.keys()

        if len(enum) > 1:
            constraint = """"%s" in ('%s')""" % (
                self.fieldname, "', '".join(enum))
        elif enum:
            constraint = """"%s" = '%s'""" % (self.fieldname, list(enum)[0])
        else:
            constraint = None

        if constraint:
            enum = list(enum)
            enum.sort()
            key = md5()
            key.update(str(enum).encode('utf-8'))
            name = self.fieldname + '_' + key.hexdigest() + '_types'
            return [CheckConstraint(constraint, name=name)]

        return []


class Json(Column):
    """ JSON column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Json


        @Declarations.register(Declarations.Model)
        class Test:

            x = Json()

    """
    sqlalchemy_type = types.JSON(none_as_null=True)


class LargeBinary(Column):
    """ Large binary column

    ::

        from os import urandom
        from anyblok.declarations import Declarations
        from anyblok.column import LargeBinary


        blob = urandom(100000)


        @Declarations.register(Declarations.Model)
        class Test:

            x = LargeBinary(default=blob)

    """
    sqlalchemy_type = types.LargeBinary


class Sequence(String):
    """ Sequence column

    ::

        from anyblok.column import Sequence


        @Declarations.register(Declarations.Model)
        class Test:

            x = Sequence()

    """
    def __init__(self, *args, **kwargs):
        if 'foreign_key' in kwargs:
            raise FieldException("Sequence column can not define a foreign key"
                                 " %r" % kwargs['foreign_key'])
        if 'default' in kwargs:
            raise FieldException("Sequence column can not define a default "
                                 "value")
        kwargs['default'] = ColumnDefaultValue(self.wrap_default)

        self.code = kwargs.pop('code') if 'code' in kwargs else None
        self.formater = kwargs.pop(
            'formater') if 'formater' in kwargs else None

        super(Sequence, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        res = super(Sequence, self).autodoc_get_properties()
        res['formater'] = self.formater
        return res

    def wrap_default(self, registry, namespace, fieldname, properties):
        if not hasattr(registry, '_need_sequence_to_create_if_not_exist'):
            registry._need_sequence_to_create_if_not_exist = []
        elif registry._need_sequence_to_create_if_not_exist is None:
            registry._need_sequence_to_create_if_not_exist = []

        code = self.code if self.code else "%s=>%s" % (namespace, fieldname)
        registry._need_sequence_to_create_if_not_exist.append(
            {'code': code, 'formater': self.formater})

        def default_value():
            return registry.System.Sequence.nextvalBy(code=code)

        return default_value


class Color(Column):
    """Color column.
    `See coulour pakage <https://pypi.python.org/pypi/colour/>`_

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Color


        @Declarations.register(Declarations.Model)
        class Test:

            x = Color(default='green')

    """
    def __init__(self, *args, **kwargs):
        max_length = kwargs.pop('size', 20)

        if 'type_' in kwargs:
            del kwargs['type_']

        self.sqlalchemy_type = ColorType(max_length)
        super(Color, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        if isinstance(value, str):
            value = self.sqlalchemy_type.python_type(value)

        return value

    def autodoc_get_properties(self):
        res = super(Color, self).autodoc_get_properties()
        res['size'] = self.max_length
        return res


class UUID(Column):
    """ UUID column

    ::

        from anyblok.column import UUID


        @Declarations.register(Declarations.Model)
        class Test:

            x = UUID()

    """
    def __init__(self, *args, **kwargs):
        uuid_kwargs = {}
        self.binary = None
        self.native = None
        for kwarg in ('binary', 'native'):
            if kwarg in kwargs:
                uuid_kwargs[kwarg] = kwargs.pop(kwarg)
                setattr(self, kwarg, uuid_kwargs[kwarg])

        self.sqlalchemy_type = UUIDType(**uuid_kwargs)
        super(UUID, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        res = super(UUID, self).autodoc_get_properties()
        res['binary'] = self.binary
        res['native'] = self.native
        return res


class URL(Column):
    """ URL column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import URL


        @Declarations.register(Declarations.Model)
        class Test:

            x = URL(default='doc.anyblok.org')

    """
    sqlalchemy_type = URLType

    def setter_format_value(self, value):
        from furl import furl

        if value is not None:
            if isinstance(value, str):
                value = furl(value)

        return value


class PhoneNumber(Column):
    """ PhoneNumber column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import PhoneNumber


        @Declarations.register(Declarations.Model)
        class Test:

            x = PhoneNumber(default='+120012301')


    .. note:: ``phonenumbers`` >= **8.9.5** distribution is required

    """
    def __init__(self, region='FR', max_length=20, *args, **kwargs):
        self.region = region
        self.max_length = max_length
        if 'type_' in kwargs:
            del kwargs['type_']

        self.sqlalchemy_type = PhoneNumberType(
            region=region, max_length=max_length)
        super(PhoneNumber, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        if isinstance(value, str):
            value = self.sqlalchemy_type.python_type(value)

        return value

    def autodoc_get_properties(self):
        res = super(PhoneNumber, self).autodoc_get_properties()
        res['region'] = self.region
        res['max_length'] = self.max_length
        return res


class Email(Column):
    """ Email column

    ::

        from anyblok.column import Email


        @Declarations.register(Declarations.Model)
        class Test:

            x = Email()

    """
    sqlalchemy_type = EmailType

    def setter_format_value(self, value):
        if value is not None:
            return value.lower()
        return value


class CountryType(types.TypeDecorator, ScalarCoercible):
    """ Generic type for Column Country """

    impl = types.Unicode(3)
    python_type = python_pycountry_type

    def process_bind_param(self, value, dialect):
        if value and isinstance(value, self.python_type):
            return value.alpha_3

        return value

    def process_result_value(self, value, dialect):
        if value:
            return pycountry.countries.get(alpha_3=value)

        return value

    def _coerce(self, value):
        if value is not None and not isinstance(value, self.python_type):
            return pycountry.countries.get(alpha_3=value)

        return value


class Country(Column):
    """Country column.

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Country
        from pycountry import countries


        @Declarations.register(Declarations.Model)
        class Test:

            x = Country(default=countries.get(alpha_2='FR'))

    """
    sqlalchemy_type = CountryType

    def __init__(self, mode='alpha_2', *args, **kwargs):
        self.mode = mode
        if pycountry is None:
            raise FieldException(
                "'pycountry' package is required for use 'CountryType'")

        self.choices = {getattr(country, mode): country.name
                        for country in pycountry.countries}
        super(Country, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        if value and not isinstance(value, self.sqlalchemy_type.python_type):
            try:
                value = pycountry.countries.get(**{self.mode: value})
            except LookupError:
                value = pycountry.countries.lookup(value)

        return value

    def autodoc_get_properties(self):
        res = super(Country, self).autodoc_get_properties()
        res['mode'] = self.mode
        res['choices'] = self.choices
        return res

    def update_properties(self, registry, namespace, fieldname, properties):
        super(Country, self).update_properties(registry, namespace,
                                               fieldname, properties)
        self.fieldname = fieldname
        properties['add_in_table_args'].append(self)

    def update_table_args(self, Model):
        """return check constraint to limit the value"""
        enum = [country.alpha_3 for country in pycountry.countries]
        constraint = """"%s" in ('%s')""" % (self.fieldname, "', '".join(enum))
        enum.sort()
        key = md5()
        key.update(str(enum).encode('utf-8'))
        name = self.fieldname + '_' + key.hexdigest() + '_types'
        return [CheckConstraint(constraint, name=name)]



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from anyblok.column import String, Json, Boolean
from ..exceptions import ParameterException


register = Declarations.register
System = Declarations.Model.System


@register(Declarations.Model.System)
class Parameter:
    """Applications parameters.

    This Model is provided by ``anyblok-core`` to give applications a uniform
    way of specifying in-database configuration.

    It is a simple key/value representation, where values can be of any type
    that can be encoded as JSON.

    A simple access API is provided with the :meth:`get`, :meth:`set`,
    :meth:`is_exist` and further methods.
    """

    key = String(primary_key=True)
    value = Json(nullable=False)
    multi = Boolean(default=False)

    @classmethod
    def set(cls, key, value):
        """ Insert or update parameter value for a key.

        .. note:: if the key already exists, the value will be updated

        :param str key: key to save
        :param value: value to save
        """
        multi = False
        if not isinstance(value, dict):
            value = {'value': value}
        else:
            multi = True

        if cls.is_exist(key):
            param = cls.from_primary_keys(key=key)
            param.update(value=value, multi=multi)
        else:
            cls.insert(key=key, value=value, multi=multi)

    @classmethod
    def is_exist(cls, key):
        """ Check if one parameter exist for the key

        :param key: key to check
        :rtype: bool
        """
        query = cls.query().filter(cls.key == key)
        return True if query.count() else False

    @classmethod
    def get(cls, key):
        """ Return the value of the key

        :param key: key whose value to retrieve
        :return: associated value
        :rtype: anything JSON encodable
        :raises ParameterException: if the key doesn't exist.
        """
        if not cls.is_exist(key):
            raise ParameterException(
                "unexisting key %r" % key)

        param = cls.from_primary_keys(key=key)
        if param.multi:
            return param.value

        return param.value['value']

    @classmethod
    def pop(cls, key):
        """Remove the given key and return the associated value.

        :param str key: the key to remove
        :return: the value before removal
        :rtype: any JSON encodable type
        :raises ParameterException: if the key wasn't present
        """
        if not cls.is_exist(key):
            raise ParameterException(
                "unexisting key %r" % key)

        param = cls.from_primary_keys(key=key)
        if param.multi:
            res = param.value
        else:
            res = param.value['value']

        param.delete()
        return res



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.

from anyblok import Declarations
from anyblok.authorization.rule.base import deny_all


@Declarations.register(Declarations.Model)
class Authorization:
    """Namespace for models supporting authorization policies."""


class DefaultModelDeclaration:
    """Pseudo model to represent the default value."""

    __registry_name__ = None


Declarations.AuthorizationBinding(DefaultModelDeclaration, deny_all)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import TestCase, DBTestCase, LogCapture
from logging import DEBUG
from anyblok.registry import RegistryManager
from anyblok.environment import EnvironmentManager
from anyblok.model import (has_sql_fields, get_fields, ModelException,
                           has_sqlalchemy_fields)
from anyblok import Declarations
from anyblok.column import Integer, String

register = Declarations.register
unregister = Declarations.unregister
Model = Declarations.Model


class OneModel:
    __tablename__ = 'test'


class TestModel(TestCase):

    @classmethod
    def setUpClass(cls):
        super(TestModel, cls).setUpClass()
        RegistryManager.init_blok('testModel')
        EnvironmentManager.set('current_blok', 'testModel')

    @classmethod
    def tearDownClass(cls):
        super(TestModel, cls).tearDownClass()
        EnvironmentManager.set('current_blok', None)
        del RegistryManager.loaded_bloks['testModel']

    def setUp(self):
        super(TestModel, self).setUp()
        blokname = 'testModel'
        RegistryManager.loaded_bloks[blokname]['Model'] = {
            'registry_names': []}

    def assertInModel(self, *args):
        blokname = 'testModel'
        blok = RegistryManager.loaded_bloks[blokname]
        self.assertEqual(len(blok['Model']['Model.MyModel']['bases']),
                         len(args))
        for cls_ in args:
            has = cls_ in blok['Model']['Model.MyModel']['bases']
            self.assertEqual(has, True)

    def assertInRemoved(self, cls):
        core = RegistryManager.loaded_bloks['testModel']['removed']
        if cls in core:
            return True

        self.fail('Not in removed')

    def test_add_interface(self):
        register(Model, cls_=OneModel, name_='MyModel')
        self.assertEqual('Model', Model.MyModel.__declaration_type__)
        self.assertInModel(OneModel)
        dir(Declarations.Model.MyModel)

    def test_add_interface_with_decorator(self):

        @register(Model)
        class MyModel:
            pass

        self.assertEqual('Model', Model.MyModel.__declaration_type__)
        self.assertInModel(MyModel)

    def test_add_two_interface(self):

        register(Model, cls_=OneModel, name_="MyModel")

        @register(Model)
        class MyModel:
            pass

        self.assertInModel(OneModel, MyModel)

    def test_remove_interface_with_1_cls_in_registry(self):

        register(Model, cls_=OneModel, name_="MyModel")
        self.assertInModel(OneModel)
        unregister(Model.MyModel, OneModel)
        self.assertInModel(OneModel)
        self.assertInRemoved(OneModel)

    def test_remove_interface_with_2_cls_in_registry(self):

        register(Model, cls_=OneModel, name_="MyModel")

        @register(Model)
        class MyModel:
            pass

        self.assertInModel(OneModel, MyModel)
        unregister(Model.MyModel, OneModel)
        self.assertInModel(MyModel, OneModel)
        self.assertInRemoved(OneModel)


def simple_model():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()


def simple_model_with_tablename():

    @register(Model, tablename='othername')
    class Test:
        id = Integer(primary_key=True)
        name = String()


def simple_models_with_same_table():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model, tablename='test')
    class Test2:
        id = Integer(primary_key=True)
        name = String()


def simple_models_with_same_table_by_declaration_model():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model, tablename=Model.Test)
    class Test2:
        id = Integer(primary_key=True)
        name = String()


def simple_models_with_same_table_by_inherit():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2:
        id = Integer(primary_key=True)
        name = String()

    @register(Model, tablename='test')  # noqa
    class Test2:
        pass


def simple_models_with_inherit_sqlmodel():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2(Model.Test):
        pass

    @register(Model)
    class Test3(Model.Test):
        pass


def model_with_foreign_key():

    @register(Model)
    class TestFk:

        name = String(primary_key=True)

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        name = String(foreign_key=Model.TestFk.use('name'))


class TestModel2(DBTestCase):

    def test_model_is_assembled(self):
        with LogCapture('anyblok.registry', level=DEBUG) as logs:
            self.init_registry(None)
            messages = logs.get_debug_messages()
            self.assertIn("Assemble 'Model' entry", messages)

    def test_model_is_initialized(self):
        with LogCapture('anyblok.registry', DEBUG) as logs:
            self.init_registry(None)
            messages = logs.get_debug_messages()
            self.assertIn("Initialize 'Model' entry", messages)

    def check_registry(self, Model):
        t = Model.insert(name="test")
        t2 = Model.query().first()
        self.assertEqual(t2, t)

    def check_registry_same_table(self, Model1, Model2):
        t = Model1.insert(name="test")
        t2 = Model2.query().first()
        self.assertEqual(t2.name, t.name)

    def test_simple_model(self):
        registry = self.init_registry(simple_model)
        self.check_registry(registry.Test)

    def test_simple_model_with_tablename(self):
        registry = self.init_registry(simple_model_with_tablename)
        self.check_registry(registry.Test)
        self.assertEqual(registry.Test.__table__.name, 'othername')
        self.assertEqual(registry.Test.__tablename__, 'othername')

    def test_simple_models_with_same_table(self):
        registry = self.init_registry(simple_models_with_same_table)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_models_with_same_table_by_declaration_model(self):
        registry = self.init_registry(
            simple_models_with_same_table_by_declaration_model)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_models_with_same_table_by_inherit(self):
        registry = self.init_registry(simple_models_with_same_table_by_inherit)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_models_with_inherit_sqlmodel(self):
        registry = self.init_registry(simple_models_with_inherit_sqlmodel)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_model_with_wrong_column(self):
        registry = self.init_registry(simple_model)

        try:
            registry.Test.insert(name="test", other="other")
            self.fail('No error when an inexisting colomn has filled')
        except TypeError:
            pass

    def test_model_with_foreign_key(self):
        registry = self.init_registry(model_with_foreign_key)
        registry.TestFk.insert(name='test')
        self.check_registry(registry.Test)

    def test_model_with_foreign_key_with_db_column_name(self):

        def model_with_foreign_key():

            @register(Model)
            class TestFk:

                name = String(primary_key=True, db_column_name='other')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                name = String(foreign_key=Model.TestFk.use('name'))

        registry = self.init_registry(model_with_foreign_key)
        registry.TestFk.insert(name='test')
        self.check_registry(registry.Test)

    def test_table_args(self):
        val = 'one arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_table_args(cls):
                    return (val,)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(len(registry.Test.__table_args__), 1)
        self.assertEqual(registry.Test.__table_args__[0], val)

    def test_table_args_with_inherit(self):
        val = 'first arg'
        val2 = 'second arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_table_args(cls):
                    return (val,)

            @register(Model)  # noqa
            class Test:

                @classmethod
                def define_table_args(cls):
                    return super(Test, cls).define_table_args() + (val2,)

        registry = self.init_registry(add_in_registry)
        self.assertEqual(len(registry.Test.__table_args__), 2)
        self.assertIn(val, registry.Test.__table_args__)
        self.assertIn(val2, registry.Test.__table_args__)

    def test_table_args_in_cls_attribute(self):

        def add_in_registry():

            @register(Model)
            class Test:
                __table_args__ = ()

        with self.assertRaises(ModelException):
            self.init_registry(add_in_registry)

    def test_mapper_args(self):
        val = 'one arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_mapper_args(cls):
                    return {val: val}

        registry = self.init_registry(add_in_registry)
        self.assertEqual(len(registry.Test.__mapper_args__), 1)
        self.assertEqual(registry.Test.__mapper_args__[val], val)

    def test_mapper_args_with_inherit(self):
        val = 'first arg'
        val2 = 'second arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_mapper_args(cls):
                    return {val: val}

            @register(Model)  # noqa
            class Test:

                @classmethod
                def define_mapper_args(cls):
                    mapper_args = super(Test, cls).define_mapper_args()
                    mapper_args.update({val2: val2})
                    return mapper_args

        registry = self.init_registry(add_in_registry)
        self.assertEqual(len(registry.Test.__mapper_args__), 2)
        self.assertEqual(registry.Test.__mapper_args__[val], val)
        self.assertEqual(registry.Test.__mapper_args__[val2], val2)

    def test_mapper_args_in_cls_attribute(self):

        def add_in_registry():

            @register(Model)
            class Test:
                __mapper_args__ = {}

        with self.assertRaises(ModelException):
            self.init_registry(add_in_registry)

    def test_with_sqlalchemy_fields(self):
        from sqlalchemy import Column as SaC, String as SaS

        def add_in_registry():

            @register(Model)
            class Test:
                one_field = SaC(SaS(64))

        with self.assertRaises(ModelException):
            self.init_registry(add_in_registry)

    def test_update_0(self):
        registry = self.init_registry(None)
        t = registry.System.Blok.query().first()
        t.update(**{'state': 'undefined'})
        self.assertEqual(t.state, 'undefined')

    def test_update_1(self):
        registry = self.init_registry(None)
        t = registry.System.Blok.query().first()
        t.update(state='undefined')
        self.assertEqual(t.state, 'undefined')

    def test_update_2(self):
        registry = self.init_registry(None)
        t = registry.System.Blok.query().first()
        t.update(state='undefined')
        self.assertEqual(t.state, 'undefined')


class TestModelAssembling(TestCase):

    def test_has_sql_fields_ok(self):

        class MyModel:
            one_field = String()

        self.assertEqual(has_sql_fields([MyModel]), True)

    def test_has_sql_fields_ko(self):

        class MyModel:
            one_field = None

        self.assertEqual(has_sql_fields([MyModel]), False)

    def test_get_fields(self):

        class MyModel:
            one_field = String()

        self.assertEqual(get_fields(MyModel), {'one_field': MyModel.one_field})

    def test_has_sqlalchemy_fields(self):
        from sqlalchemy import Column as SaC, String as SaS

        class MyModel:
            one_field = SaC(SaS(64))

        self.assertTrue(has_sqlalchemy_fields(MyModel))

    def test_has_sqlalchemy_fields2(self):

        class MyModel:
            one_field = String()

        self.assertFalse(has_sqlalchemy_fields(MyModel))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Georges Racinet <gracinet@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .base import AuthorizationRule
from .base import RuleNotForModelClasses


class AttributeAccessRule(AuthorizationRule):
    """Grant authorization to principals coinciding with a record attribute.

    Whatever the permission is associated to this policy, it will be granted
    to principals on records whose attribute is equal to the principal.

    A common use-case is to associate it to a precise permission, in
    conjunction with a flatter default policy, such as
    :class:`..model_authz.ModelBasedAutorizationRule`
    """

    def __init__(self, attr, model_rule=None):
        """.

        :param attr: The attribute that is being compared with principal.
        :param model_rule: If set, checks done on model classes will be
                           relayed to this other rule. Otherwise, the
                           standard exception that the rule does not apply
                           to model classes is raised.

        The ``model_rule`` allows to express conveniently that some
        principal has a "general" Read right on some model,
        while still allowing it to read only some of the
        records, and to protect the querying by the same
        'Read' permission. Similar and finer effects can
        be obtained by creating a separate 'Search' permission, but that may
        not be appropriate in a given context.
        """
        self.attr = attr
        self.model_rule = model_rule

        self._registry = None

    @property
    def registry(self):
        """On this rule, we'll need a setter for registry"""
        return self._registry

    @registry.setter
    def registry(self, registry):
        """Apply registry also to model_rule if needed"""
        self._registry = registry
        if self.model_rule is not None:
            self.model_rule.registry = registry

    def check(self, record, principals, permission):
        if isinstance(record, type):
            if self.model_rule is not None:
                return self.model_rule.check(record, principals, permission)
            raise RuleNotForModelClasses(self, record)

        return getattr(record, self.attr) in principals

    def filter(self, model, query, principals, permission):
        return query.filter(getattr(model, self.attr).in_(principals))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from json import dumps


class TestImporterFormater(BlokTestCase):

    def get_value(self, value, ctype, external_id=False, model=None):
        return self.registry.IO.Importer().str2value(value, ctype,
                                                     external_id=external_id,
                                                     model=model)

    def test_datetime(self):
        from datetime import datetime
        value = self.get_value("2015-05-01 17:52:01", "DateTime")
        self.assertTrue(isinstance(value, datetime))
        self.assertEqual(value.year, 2015)
        self.assertEqual(value.month, 5)
        self.assertEqual(value.day, 1)

    def test_decimal(self):
        from decimal import Decimal
        value = self.get_value("1.5", "Decimal")
        self.assertTrue(isinstance(value, Decimal))
        self.assertEqual(value, Decimal("1.5"))

    def test_float(self):
        value = self.get_value("1.5", "Float")
        self.assertTrue(isinstance(value, float))
        self.assertEqual(value, 1.5)

    def test_json(self):
        value = self.get_value(dumps({'id': 1}), "Json")
        self.assertTrue(isinstance(value, dict))
        self.assertEqual(value, {'id': 1})

    def test_time(self):
        from datetime import time
        value = self.get_value("17:52:01", "Time")
        self.assertTrue(isinstance(value, time))
        self.assertEqual(value.hour, 17)

    def test_big_interger(self):
        value = self.get_value("17", "BigInteger")
        self.assertTrue(isinstance(value, int))
        self.assertEqual(value, 17)

    def test_boolean(self):
        for v in ('1', 'true', 'True'):
            value = self.get_value(v, "Boolean")
            self.assertTrue(isinstance(value, bool))
            self.assertEqual(value, True)

        for v in ('0', 'false', 'False'):
            value = self.get_value(v, "Boolean")
            self.assertTrue(isinstance(value, bool))
            self.assertEqual(value, False)

    def test_date(self):
        from datetime import date
        value = self.get_value("2015-05-01", "Date")
        self.assertTrue(isinstance(value, date))
        self.assertEqual(value.year, 2015)
        self.assertEqual(value.month, 5)
        self.assertEqual(value.day, 1)

    def test_interger(self):
        value = self.get_value("10", "Integer")
        self.assertTrue(isinstance(value, int))
        self.assertEqual(value, 10)

    def test_interval(self):
        from datetime import timedelta
        value = self.get_value("10", "Interval")
        self.assertTrue(isinstance(value, timedelta))
        self.assertEqual(value.seconds, 10)

    def test_large_binary(self):
        from os import urandom
        blob = urandom(100)
        value = self.get_value(blob, "Large_Binary")
        self.assertTrue(isinstance(value, bytes))
        self.assertEqual(value, blob)

    def test_selection(self):
        value = self.get_value("selection", "Selection")
        self.assertTrue(isinstance(value, str))
        self.assertEqual(value, "selection")

    def test_string(self):
        value = self.get_value("selection", "String")
        self.assertTrue(isinstance(value, str))
        self.assertEqual(value, "selection")

    def test_text(self):
        value = self.get_value("selection", "Text")
        self.assertTrue(isinstance(value, str))
        self.assertEqual(value, "selection")

    def test_many2many(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value(
            dumps([pks]), "Many2Many", model="Model.System.Model")
        self.assertEqual(value, [model])

    def test_many2many_multi_value(self):
        pks1 = dict(name='Model.System.Model')
        pks2 = dict(name='Model.System.Column')
        model1 = self.registry.System.Model.from_primary_keys(**pks1)
        model2 = self.registry.System.Model.from_primary_keys(**pks2)
        value = self.get_value(
            dumps([pks1, pks2]), "Many2Many", model="Model.System.Model")
        self.assertEqual(value, [model1, model2])

    def test_many2many_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(dumps([key]), "Many2Many",
                               external_id=True, model="Model.System.Model")
        self.assertEqual(value, [model])

    def test_many2many_multi_value_external_ids(self):
        key1 = 'formater_mapping1'
        key2 = 'formater_mapping2'
        model1 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        model2 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Column')
        self.registry.IO.Mapping.set(key1, model1)
        self.registry.IO.Mapping.set(key2, model2)
        value = self.get_value(dumps([key1, key2]), "Many2Many",
                               external_id=True, model="Model.System.Model")
        self.assertEqual(value, [model1, model2])

    def test_one2many(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value(
            dumps([pks]), "One2Many", model="Model.System.Model")
        self.assertEqual(value, [model])

    def test_one2many_multi_value(self):
        pks1 = dict(name='Model.System.Model')
        pks2 = dict(name='Model.System.Column')
        model1 = self.registry.System.Model.from_primary_keys(**pks1)
        model2 = self.registry.System.Model.from_primary_keys(**pks2)
        value = self.get_value(
            dumps([pks1, pks2]), "One2Many", model="Model.System.Model")
        self.assertEqual(value, [model1, model2])

    def test_one2many_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(dumps([key]), "One2Many", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, [model])

    def test_one2many_external_ids_is_none_1(self):
        value = self.get_value(dumps(None), "One2Many", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, [])

    def test_one2many_external_ids_is_none_2(self):
        value = self.get_value(dumps([]), "One2Many", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, [])

    def test_one2many_multi_value_external_ids(self):
        key1 = 'formater_mapping1'
        key2 = 'formater_mapping2'
        model1 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        model2 = self.registry.System.Model.from_primary_keys(
            name='Model.System.Column')
        self.registry.IO.Mapping.set(key1, model1)
        self.registry.IO.Mapping.set(key2, model2)
        value = self.get_value(dumps([key1, key2]), "One2Many",
                               external_id=True, model="Model.System.Model")
        self.assertEqual(value, [model1, model2])

    def test_many2one(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value(
            dumps(pks), "Many2One", model="Model.System.Model")
        self.assertEqual(value, model)

    def test_many2one_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(key, "Many2One", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, model)

    def test_many2one_external_ids_with_none_value(self):
        value = self.get_value(None, "Many2One", external_id=True,
                               model="Model.System.Model")
        self.assertIsNone(value)

    def test_one2one(self):
        pks = dict(name='Model.System.Model')
        model = self.registry.System.Model.from_primary_keys(**pks)
        value = self.get_value(
            dumps(pks), "One2One", model="Model.System.Model")
        self.assertEqual(value, model)

    def test_one2one_external_ids(self):
        key = 'formater_mapping'
        model = self.registry.System.Model.from_primary_keys(
            name='Model.System.Model')
        self.registry.IO.Mapping.set(key, model)
        value = self.get_value(key, "One2One", external_id=True,
                               model="Model.System.Model")
        self.assertEqual(value, model)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations
from sqlalchemy import Sequence as SQLASequence
from anyblok.column import Integer, String


register = Declarations.register
System = Declarations.Model.System


@register(System)
class Sequence:
    """Database sequences.

    This Model allows applications to define and use Database sequences easily.

    It is a rewrapping of `SQLAlchemy sequences
    <http://docs.sqlalchemy.org/en/latest/core/defaults.html
    #sqlalchemy.schema.Sequence>`_, with additional formatting
    capabilities to use them, e.g, in fields of applicative Models.

    Sample usage::

        sequence = registry.System.Sequence.insert(
        code="string code",
        formater="One prefix {seq} One suffix")

    .. seealso:: The :attr:`formater` field.

    To get the next formatted value of the sequence::

        sequence.nextval()

    Full example in a Python shell::

        >>> seq = Sequence.insert(code='SO', formater="{code}-{seq:06d}")
        >>> seq.nextval()
        'SO-000001'
        >>> seq.nextval()
        'SO-000002'
    """

    _cls_seq_name = 'system_sequence_seq_name'

    id = Integer(primary_key=True)
    code = String(nullable=False)
    number = Integer(nullable=False)
    seq_name = String(nullable=False)
    """Name of the sequence in the database.

    Most databases identify sequences by names which must be globally
    unique.

    If not passed at insertion, the value of this field is automatically
    generated.
    """
    formater = String(nullable=False, default="{seq}")
    """Python format string to render the sequence values.

    This format string is used in :meth:`nextval`. Within it, you can use the
    following variables:

       * seq: current value of the underlying database sequence
       * code: :attr:`code` field
       * id: :attr:`id` field
    """

    @classmethod
    def initialize_model(cls):
        """ Create the sequence to determine name """
        super(Sequence, cls).initialize_model()
        seq = SQLASequence(cls._cls_seq_name)
        seq.create(cls.registry.bind)

        to_create = getattr(cls.registry,
                            '_need_sequence_to_create_if_not_exist', ())
        if to_create is None:
            return

        for vals in to_create:
            if cls.query().filter(cls.code == vals['code']).count():
                continue

            formatter = vals.get('formater')
            if formatter is None:
                del vals['formater']

            cls.insert(**vals)

    @classmethod
    def create_sequence(cls, values):
        """Create the database sequence for an instance of Sequence Model.

        :return: suitable field values for insertion of the Model instance
        :rtype: dict
        """
        seq_name = values.get('seq_name')
        if seq_name is None:
            seq_id = cls.registry.execute(SQLASequence(cls._cls_seq_name))
            seq_name = '%s_%d' % (cls.__tablename__, seq_id)
            values['seq_name'] = seq_name

        number = values.setdefault('number', 0)
        if number:
            seq = SQLASequence(seq_name, number)
        else:
            seq = SQLASequence(seq_name)
        seq.create(cls.registry.bind)
        return values

    @classmethod
    def insert(cls, **kwargs):
        """Overwrite to call :meth:`create_sequence` on the fly."""
        return super(Sequence, cls).insert(**cls.create_sequence(kwargs))

    @classmethod
    def multi_insert(cls, *args):
        """Overwrite to call :meth:`create_sequence` on the fly."""
        res = [cls.create_sequence(x) for x in args]
        return super(Sequence, cls).multi_insert(*res)

    def nextval(self):
        """Format and return the next value of the sequence.

        :rtype: str
        """
        nextval = self.registry.execute(SQLASequence(self.seq_name))
        self.update(number=nextval)
        return self.formater.format(code=self.code, seq=nextval, id=self.id)

    @classmethod
    def nextvalBy(cls, **crit):
        """Return next value of the first Sequence matching given criteria.

        :param crit: criteria to match, e.g., ``code=SO``
        :return: :meth:`next_val` result for the first matching Sequence,
                 or ``None`` if there's no match.
        """
        filters = [getattr(cls, k) == v for k, v in crit.items()]
        seq = cls.query().filter(*filters).first()
        if seq is None:
            return None
        return seq.nextval()



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase
from anyblok import Declarations
from anyblok.mapper import ModelAttributeException
from anyblok.column import Integer, String, DateTime
from anyblok.relationship import Many2Many, Many2One
from anyblok.field import FieldException
from datetime import datetime

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


def _complete_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


def _minimum_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def _minimum_many2many_by_mixin(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Mixin)
    class MixinM2M:
        addresses = Many2Many(model=Model.Address)

    @register(Model)
    class Person(Mixin.MixinM2M):

        name = String(primary_key=True)


def _many2many_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model='Model.Address')


def auto_detect_two_primary_keys(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        id2 = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def unexisting_remote_columns(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address, remote_columns="id2")


def reuse_many2many_table(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)

    @register(Model)  # noqa
    class Address:

        id = Integer(primary_key=True)
        persons = Many2Many(model=Model.Person,
                            join_table='join_person_and_address')


class TestMany2Many(DBTestCase):

    def test_complete_many2many(self):
        registry = self.init_registry(_complete_many2many)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_addresses_by_persons' in jt
        self.assertEqual(join_table_exist, True)

        self.assertEqual(
            len(jt['join_addresses_by_persons'].primary_key.columns),
            2)
        self.assertTrue('a_id' in jt[
            'join_addresses_by_persons'].primary_key.columns)
        self.assertTrue('p_name' in jt[
            'join_addresses_by_persons'].primary_key.columns)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(address.persons, [person])

    def test_complete_many2many_expire_field(self):
        registry = self.init_registry(_complete_many2many)
        self.assertIn(
            ('x2m', 'addresses', 'persons'),
            registry.expire_attributes['Model.Person']['addresses'])
        self.assertIn(
            ('x2m', 'persons', 'addresses'),
            registry.expire_attributes['Model.Address']['persons'])

    def test_minimum_many2many(self):
        registry = self.init_registry(_minimum_many2many)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(person.addresses, [address])

    def test_many2many_with_str_model(self):
        registry = self.init_registry(_many2many_with_str_model)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(person.addresses, [address])

    def test_unexisting_remote_columns(self):
        with self.assertRaises(ModelAttributeException):
            self.init_registry(unexisting_remote_columns)

    def test_reuse_many2many_table(self):
        registry = self.init_registry(reuse_many2many_table)

        address = registry.Address.insert()
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(address.persons, [person])

    def test_declared_in_mixin(self):
        registry = self.init_registry(_minimum_many2many_by_mixin)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        self.assertEqual(person.addresses, [address])

    def test_declared_in_mixin_inherit_by_two_models(self):
        def add_in_registry():
            _minimum_many2many_by_mixin()

            @register(Model)
            class Person2(Mixin.MixinM2M):

                name = String(primary_key=True)

        registry = self.init_registry(add_in_registry)

        address_exist = hasattr(registry.Person, 'addresses')
        self.assertTrue(address_exist)
        address_exist = hasattr(registry.Person2, 'addresses')
        self.assertTrue(address_exist)

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        self.assertTrue(m2m_tables_exist)

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address' in jt
        self.assertEqual(join_table_exist, True)
        join_table_exist = 'join_person2_and_address' in jt
        self.assertEqual(join_table_exist, True)

    def test_comlet_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

        jt = registry.declarativebase.metadata.tables
        self.assertEqual(
            len(jt['join_test_and_test2'].primary_key.columns),
            4)
        self.assertTrue('t1_id' in jt[
            'join_test_and_test2'].primary_key.columns)
        self.assertTrue('t1_id2' in jt[
            'join_test_and_test2'].primary_key.columns)
        self.assertTrue('t2_id' in jt[
            'join_test_and_test2'].primary_key.columns)
        self.assertTrue('t2_id2' in jt[
            'join_test_and_test2'].primary_key.columns)

    def test_comlet_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns='id',
                                 m2m_local_columns='t1_id',
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_comlet_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns='id',
                                 m2m_remote_columns='t2_id',
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2='test2')
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_minimum_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_minimum_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_minimum_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        self.assertIs(t1.test2[0], t2)

    def test_many2many_on_self(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    m2m_remote_columns='id2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.childs.append(t2)
        self.assertIn(t1, t2.parents)

    def test_many2many_on_self_auto_column(self):

        def add_in_registry():

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test2.insert()
        t2 = registry.Test2.insert()
        t1.childs.append(t2)
        self.assertIn(t1, t2.parents)

    def test_many2many_on_multi_fk_miss_on_m2m_column(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(
                    m2m_remote_columns=['test_id'],
                    model=Model.Address)

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_complete_config(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="personaddress",
                                      remote_columns="id", local_columns="name",
                                      m2m_remote_columns='a_id',
                                      m2m_local_columns='p_name',
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        self.assertEqual(personaddress.a_id, address.id)
        self.assertEqual(personaddress.p_name, person.name)
        self.assertTrue(personaddress.id)
        self.assertTrue(personaddress.create_at)
        self.assertEqual(personaddress.foo, 'bar')

    def test_rich_many2many_minimum_config(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="personaddress",
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        self.assertEqual(personaddress.a_id, address.id)
        self.assertEqual(personaddress.p_name, person.name)
        self.assertTrue(personaddress.id)
        self.assertTrue(personaddress.create_at)
        self.assertEqual(personaddress.foo, 'bar')

    def test_rich_many2many_minimum_config_on_join_model(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        self.assertEqual(personaddress.a_id, address.id)
        self.assertEqual(personaddress.p_name, person.name)
        self.assertTrue(personaddress.id)
        self.assertTrue(personaddress.create_at)
        self.assertEqual(personaddress.foo, 'bar')

    def test_rich_many2many_minimum_config_with_many2one(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                person = Many2One(
                    model='Model.Person', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                address = Many2One(
                    model=Model.Address, nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        self.assertEqual(personaddress.address_id, address.id)
        self.assertEqual(personaddress.person_name, person.name)
        self.assertTrue(personaddress.id)
        self.assertTrue(personaddress.create_at)
        self.assertEqual(personaddress.foo, 'bar')

    def test_rich_many2many_minimum_config_with_pk_many2one(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                person = Many2One(
                    model='Model.Person', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                address = Many2One(
                    model=Model.Address, nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        self.assertEqual(personaddress.address_id, address.id)
        self.assertEqual(personaddress.person_name, person.name)
        self.assertTrue(personaddress.create_at)
        self.assertEqual(personaddress.foo, 'bar')

    def test_rich_many2many_minimum_config_on_join_model_and_join_table_1(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        self.assertEqual(personaddress.a_id, address.id)
        self.assertEqual(personaddress.p_name, person.name)
        self.assertTrue(personaddress.id)
        self.assertTrue(personaddress.create_at)
        self.assertEqual(personaddress.foo, 'bar')

    def test_rich_many2many_minimum_config_on_join_model_and_join_table_2(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="other_join_table",
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_complete_config_on_self(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    remote_columns="id", local_columns="id",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        self.assertEqual(link.t_left, t2.id)
        self.assertEqual(link.t_right, t1.id)
        self.assertTrue(link.id)
        self.assertTrue(link.create_at)
        self.assertEqual(link.foo, 'bar')

    def test_rich_many2many_minimum_config_on_self(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        self.assertEqual(link.t_left, t2.id)
        self.assertEqual(link.t_right, t1.id)
        self.assertTrue(link.id)
        self.assertTrue(link.create_at)
        self.assertEqual(link.foo, 'bar')

    def test_rich_many2many_minimum_config_on_self_without_columns_1(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                )

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_minimum_config_on_self_without_columns_2(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_remote_columns='t_right',
                )

        with self.assertRaises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_minimum_config_on_self_with_join_model(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        self.assertEqual(link.t_left, t2.id)
        self.assertEqual(link.t_right, t1.id)
        self.assertTrue(link.id)
        self.assertTrue(link.create_at)
        self.assertEqual(link.foo, 'bar')

    def test_rich_many2many_minimum_config_on_self_with_many2one(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                left = Many2One(
                    model='Model.Test', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                right = Many2One(
                    model='Model.Test', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='left_id',
                    m2m_remote_columns='right_id',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        self.assertEqual(link.left_id, t2.id)
        self.assertEqual(link.right_id, t1.id)
        self.assertTrue(link.id)
        self.assertTrue(link.create_at)
        self.assertEqual(link.foo, 'bar')

    def test_rich_many2many_minimum_config_on_self_with_pk_many2one_2(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                left = Many2One(
                    model='Model.Test', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                right = Many2One(
                    model='Model.Test', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='left',
                    m2m_remote_columns='right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        self.assertEqual(link.left_id, t2.id)
        self.assertEqual(link.right_id, t1.id)
        self.assertTrue(link.create_at)
        self.assertEqual(link.foo, 'bar')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .exceptions import (
    ForbidUpdateException, ForbidDeleteException)
from anyblok import Declarations
from anyblok.column import Boolean, Selection

Mixin = Declarations.Mixin


@Declarations.register(Mixin)
class ForbidUpdate:

    @classmethod
    def before_update_orm_event(cls, mapper, connection, target):
        modified_fields = target.get_modified_fields()
        raise ForbidUpdateException(
            "The modification of %r on %s with field(s) %r is forbidden" % (
                target, cls.__registry_name__, modified_fields.keys()))


@Declarations.register(Mixin)
class ForbidDelete:

    @classmethod
    def before_delete_orm_event(cls, mapper, connection, target):
        raise ForbidDeleteException(
            "The deletion of %r on %r is forbidden" % (
                target, cls.__registry_name__))


@Declarations.register(Declarations.Mixin)
class ReadOnly(Mixin.ForbidUpdate, Mixin.ForbidDelete):
    pass


@Declarations.register(Mixin)
class ConditionalForbidUpdate:

    def check_if_forbid_update_condition_is_true(self, **previous_values):
        raise ForbidUpdateException(
            "No 'check_if_forbid_update_condition_is_true' method found "
            " on %s to check the allow or forbid choice " % (
                self.__registry_name__))

    @classmethod
    def before_update_orm_event(cls, mapper, connection, target):
        modified_fields = target.get_modified_fields()
        if target.check_if_forbid_update_condition_is_true(**modified_fields):
            raise ForbidUpdateException(
                "The modification of %r on %s with field(s) %r is forbidden" % (
                    target, cls.__registry_name__, modified_fields.keys()))


@Declarations.register(Mixin)
class ConditionalForbidDelete:

    def check_if_forbid_delete_condition_is_true(self):
        raise ForbidDeleteException(
            "No 'check_if_forbid_delete_condition_is_true' method found "
            " on %s to check the allow or forbid choice " % (
                self.__registry_name__))

    @classmethod
    def before_delete_orm_event(cls, mapper, connection, target):
        if target.check_if_forbid_delete_condition_is_true():
            raise ForbidDeleteException(
                "The deletion of %r on %r is forbidden" % (
                    target, cls.__registry_name__))


@Declarations.register(Declarations.Mixin)
class ConditionalReadOnly(Mixin.ConditionalForbidUpdate,
                          Mixin.ConditionalForbidDelete):
    pass


@Declarations.register(Declarations.Mixin)
class BooleanForbidUpdate(Mixin.ConditionalForbidUpdate):

    forbid_update = Boolean(default=False)

    def check_if_forbid_update_condition_is_true(self, **previous_values):
        return previous_values.get('forbid_update', self.forbid_update)


@Declarations.register(Declarations.Mixin)
class BooleanForbidDelete(Mixin.ConditionalForbidDelete):

    forbid_delete = Boolean(default=False)

    def check_if_forbid_delete_condition_is_true(self):
        return self.forbid_delete


@Declarations.register(Declarations.Mixin)
class BooleanReadOnly(Mixin.ConditionalForbidUpdate,
                      Mixin.ConditionalForbidDelete):

    readonly = Boolean(default=False)

    def check_if_forbid_update_condition_is_true(self, **previous_values):
        return previous_values.get('readonly', self.readonly)

    def check_if_forbid_delete_condition_is_true(self):
        return self.readonly


@Declarations.register(Declarations.Mixin)
class StateReadOnly(Mixin.ConditionalForbidUpdate,
                    Mixin.ConditionalForbidDelete):
    DEFAULT_STATE = None

    state = Selection(selections='get_states',
                      default='get_default_state',
                      nullable=False)

    @classmethod
    def get_default_state(cls):
        return cls.DEFAULT_STATE

    @classmethod
    def get_states(cls):
        return {}



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.registry import RegistryManager
from anyblok import Declarations


@Declarations.add_declaration_type()
class Core:
    """ The Core class is the base of all the AnyBlok models

    Add new core model::

        @Declarations.register(Declarations.Core)
        class Base:
            pass

    Remove the core model::

        Declarations.unregister(Declarations.Core, 'Base', Base,
                                     blok='MyBlok')

    """

    @classmethod
    def register(self, parent, name, cls_, **kwargs):
        """ Add new sub registry in the registry

        :param parent: Existing declaration
        :param name: Name of the new declaration to add it
        :param cls_: Class Interface to add in the declaration
        """
        if not hasattr(parent, name):
            core = type(name, tuple(), {})
            setattr(parent, name, core)

        if parent == Declarations:
            return

        RegistryManager.add_core_in_register(name, cls_)

    @classmethod
    def unregister(self, entry, cls_):
        """ Remove the Interface from the registry

        :param entry: entry declaration of the model where the ``cls_``
            must be removed
        :param cls_: Class Interface to remove in the declaration
        """
        RegistryManager.remove_in_register(cls_)


RegistryManager.declare_core('Base')
RegistryManager.declare_core('SqlBase')
RegistryManager.declare_core('SqlViewBase')
RegistryManager.declare_core('Session')
RegistryManager.declare_core('Query')
RegistryManager.declare_core('InstrumentedList')



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from .plugins import ModelPluginBase
from .exceptions import ModelException
from sqlalchemy import ForeignKeyConstraint, CheckConstraint
from sqlalchemy.ext.declarative import declared_attr


class TableMapperPlugin(ModelPluginBase):

    def initialisation_tranformation_properties(self, properties,
                                                transformation_properties):
        """ Initialise the transform properties: hybrid_method

        :param new_type_properties: param to add in a new base if need
        """
        properties['add_in_table_args'] = []
        if 'table_args' not in transformation_properties:
            transformation_properties['table_args'] = False
        if 'mapper_args' not in transformation_properties:
            transformation_properties['mapper_args'] = False

    def transform_base(self, namespace, base,
                       transformation_properties,
                       new_type_properties):
        """Test if define_table/mapper_args are in the base, and call them
        save the value in the properties

        if  __table/mapper_args\_\_ are in the base then raise ModelException

        :param namespace: the namespace of the model
        :param base: One of the base of the model
        :param transformation_properties: the properties of the model
        :param new_type_properties: param to add in a new base if need
        """
        if hasattr(base, '__table_args__'):
            raise ModelException(
                "'__table_args__' attribute is forbidden, on Model : %r (%r)."
                "Use the class method 'define_table_args' to define the value "
                "allow anyblok to fill his own '__table_args__' attribute" % (
                    namespace, base.__table_args__))

        if hasattr(base, '__mapper_args__'):
            raise ModelException(
                "'__mapper_args__' attribute is forbidden, on Model : %r (%r)."
                "Use the class method 'define_mapper_args' to define the "
                "value allow anyblok to fill his own '__mapper_args__' "
                "attribute" % (namespace, base.__mapper_args__))

        if hasattr(base, 'define_table_args'):
            transformation_properties['table_args'] = True

        if hasattr(base, 'define_mapper_args'):
            transformation_properties['mapper_args'] = True

    def insert_in_bases(self, new_base, namespace, properties,
                        transformation_properties):
        """ Create overwrite to define table and mapper args to define some
        options for SQLAlchemy

        :param new_base: the base to be put on front of all bases
        :param namespace: the namespace of the model
        :param properties: the properties declared in the model
        :param transformation_properties: the properties of the model
        """
        table_args = tuple(properties['add_in_table_args'])
        if table_args:
            def define_table_args(cls_):
                if cls_.__registry_name__ == namespace:
                    res = super(new_base, cls_).define_table_args()
                    fks = [x.name for x in res
                           if isinstance(x, ForeignKeyConstraint)]

                    t_args = []
                    for field in table_args:
                        for constraint in field.update_table_args(cls_):
                            if (
                                not isinstance(constraint,
                                               ForeignKeyConstraint) or
                                constraint.name not in fks
                            ):
                                t_args.append(constraint)
                            elif isinstance(constraint, CheckConstraint):
                                t_args.append(constraint)

                    return res + tuple(t_args)

                return ()

            new_base.define_table_args = classmethod(define_table_args)
            transformation_properties['table_args'] = True

        self.insert_in_bases_table_args(new_base, transformation_properties)
        self.insert_in_bases_mapper_args(new_base, transformation_properties)

    def insert_in_bases_table_args(self, new_base, transformation_properties):
        """Add table __table_args__ in new_base

        :param new_base: the base to be put on front of all bases
        :param transformation_properties: the properties of the model
        """
        if transformation_properties['table_args']:

            def __table_args__(cls_):
                return cls_.define_table_args()

            new_base.__table_args__ = declared_attr(__table_args__)

    def insert_in_bases_mapper_args(self, new_base, transformation_properties):
        """Add table __mapper_args__ in new_base

        :param new_base: the base to be put on front of all bases
        :param transformation_properties: the properties of the model
        """
        if transformation_properties['mapper_args']:

            def __mapper_args__(cls_):
                return cls_.define_mapper_args()

            new_base.__mapper_args__ = declared_attr(__mapper_args__)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from nose.plugins import Plugin
from anyblok.config import Configuration, get_db_name
from anyblok import (
    load_init_function_from_entry_points,
    configuration_post_load,
)
from anyblok.blok import BlokManager
from anyblok.registry import RegistryManager, return_list
from os.path import join
from os.path import relpath
from os.path import normpath
from os import pardir
from os import walk


def isindir(path, dirpath):
    # normpath simplifies stuff like a/../c but doesn't follow symlinks
    # that's what we need. Nose will feed us absolute paths, btw
    return not relpath(normpath(path), normpath(dirpath)).startswith(pardir)


class Arg2OptOptions:

    def __init__(self, options):
        self.options = options

    def _get_args(self):
        return False

    def _get_kwargs(self):
        keys = ['configfile', 'db_name', 'db_driver_name', 'db_user_name',
                'db_password', 'db_host', 'db_port']

        return [(x, getattr(self.options, x))
                for x in keys] + [('withoutautomigration', True)]

    def __getattr__(self, name, default=None):
        return getattr(self.options, name, default)


class AnyBlokPlugin(Plugin):
    name = 'anyblok-bloks'
    score = 100

    def __init__(self):
        super(AnyBlokPlugin, self).__init__()
        self.registryLoaded = False
        self.AnyBlokOptions = None

    def options(self, parser, env):
        super(AnyBlokPlugin, self).options(parser, env)
        parser.add_option("--anyblok-configfile", dest="configfile")
        parser.add_option('--anyblok-db-name', dest='db_name',
                          default=env.get('ANYBLOK_DATABASE_NAME'),
                          help="Name of the database")
        parser.add_option('--anyblok-db-driver-name', dest='db_driver_name',
                          default=env.get('ANYBLOK_DATABASE_DRIVER'),
                          help="the name of the database backend. This name "
                               "will correspond to a module in "
                               "sqlalchemy/databases or a third party plug-in")
        parser.add_option('--anyblok-db-user-name', dest='db_user_name',
                          default=env.get('ANYBLOK_DATABASE_USER'),
                          help="The user name")
        parser.add_option('--anyblok-db-password', dest='db_password',
                          default=env.get('ANYBLOK_DATABASE_PASSWORD'),
                          help="database password")
        parser.add_option('--anyblok-db-host', dest='db_host',
                          default=env.get('ANYBLOK_DATABASE_HOST'),
                          help="The name of the host")
        parser.add_option('--anyblok-db-port', dest='db_port',
                          default=env.get('ANYBLOK_DATABASE_PORT'),
                          help="The port number")
        parser.add_option('--anyblok-db-url', dest='db_url',
                          default=env.get('ANYBLOK_DATABASE_URL'),
                          help="Complete URL for connection with the database")

    def configure(self, options, conf):
        super(AnyBlokPlugin, self).configure(options, conf)
        if self.enabled:
            self.AnyBlokOptions = Arg2OptOptions(options)

    def load_registry(self):
        if not self.enabled or self.registryLoaded:
            return

        # Load the registry here not in configuration,
        # because the configurations are not loaded in order of score
        self.registryLoaded = True
        load_init_function_from_entry_points(unittest=True)
        Configuration.load_config_for_test()
        Configuration.parse_options(self.AnyBlokOptions)
        configuration_post_load()
        BlokManager.load()
        db_name = get_db_name()

        registry = RegistryManager.get(db_name)
        if not registry:
            return

        installed_bloks = registry.System.Blok.list_by_state("installed")
        selected_bloks = return_list(
            Configuration.get('selected_bloks')) or installed_bloks

        unwanted_bloks = return_list(Configuration.get('unwanted_bloks')) or []

        self.authorized_blok_paths = set(
            BlokManager.getPath(b) for b in BlokManager.list()
            if b in selected_bloks and b not in unwanted_bloks)

        test_dirs = self.authorized_blok_test_dirs = set()
        for startpath in self.authorized_blok_paths:
            for root, dirs, _ in walk(startpath):
                if 'tests' in dirs:
                    test_dirs.add(join(root, 'tests'))

        registry.close()  # free the registry to force create it again

    def file_from_authorized_blok_tests(self, file_path):
        return any(isindir(file_path, tp)
                   for tp in self.authorized_blok_test_dirs)

    def wantModule(self, module):
        self.load_registry()
        return True

    def wantFile(self, file_path, package=None):
        self.load_registry()
        return (self.enabled and
                file_path.endswith(".py") and
                self.file_from_authorized_blok_tests(file_path))

    def wantDirectory(self, path):
        self.load_registry()
        return (self.enabled and
                any(isindir(path, bp) for bp in self.authorized_blok_paths))



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.registry import RegistryManager
from anyblok import Declarations


class MixinType:

    @classmethod
    def register(cls, parent, name, cls_, **kwargs):
        """ add new sub registry in the registry

        :param parent: Existing global registry
        :param name: Name of the new registry to add it
        :param cls_: Class Interface to add in registry
        """
        _registryname = parent.__registry_name__ + '.' + name

        if not hasattr(parent, name):
            kwargs['__registry_name__'] = _registryname
            ns = type(name, tuple(), kwargs)
            setattr(parent, name, ns)

        if parent is Declarations:
            return

        RegistryManager.add_entry_in_register(
            cls.__name__, _registryname, cls_, **kwargs)

    @classmethod
    def unregister(cls, entry, cls_):
        """ Remove the Interface in the registry

        :param entry: entry declaration of the model where the ``cls_``
            must be removed
        :param cls_: Class Interface to remove in registry
        """
        RegistryManager.remove_in_register(cls_)


@Declarations.add_declaration_type(isAnEntry=True)
class Mixin(MixinType):
    """ The Mixin class are used to define a behaviours on models:

    * Add new mixin class::

        @Declarations.register(Declarations.Mixin)
        class MyMixinclass:
            pass

    * Remove a mixin class::

        Declarations.unregister(Declarations.Mixin.MyMixinclass, MyMixinclass)
    """

    autodoc_anyblok_bases = True

    autodoc_anyblok_fields = True



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.ext.hybrid import hybrid_property
from anyblok.common import anyblok_column_prefix
from anyblok.mapper import ModelRepr


class FieldException(Exception):
    """ Simple Exception for Field """


class Field:
    """ Field class

    This class must not be instanciated
    """

    use_hybrid_property = False

    def __init__(self, *args, **kwargs):
        """ Initialize the field

        :param label: label of this field
        :type label: str
        """
        self.forbid_instance(Field)
        self.label = None

        if 'label' in kwargs:
            self.label = kwargs.pop('label')

        self.context = kwargs.pop('context', {})
        self.args = args
        self.kwargs = kwargs

    def forbid_instance(self, cls):
        """ Raise an exception if the cls is an instance of this __class__

        :param cls: instance of the class
        :exception: FieldException
        """
        if self.__class__ is cls:
            raise FieldException(
                "%r class must not be instanciated use a sub class" % cls)

    def update_description(self, registry, model, res):
        res.update(self.context)

    def update_properties(self, registry, namespace, fieldname, properties):
        """ Update the propertie use to add new column

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """

    def get_property(self, registry, namespace, fieldname, properties):
        """Return the property of the field

        .. warning::

            In the case of the get is called in classattribute,
            SQLAlchemy wrap for each call the column, the id of the wrapper
            is not the same

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """
        return hybrid_property(
            self.wrap_getter_column(fieldname),
            self.wrap_setter_column(fieldname),
            expr=self.wrap_expr_column(fieldname))

    def getter_format_value(self, value):
        return value

    def wrap_getter_column(self, fieldname):
        """Return a default getter for the field

        :param fieldname: name of the field
        """
        attr_name = anyblok_column_prefix + fieldname

        def getter_column(model_self):
            return self.getter_format_value(getattr(model_self, attr_name))

        return getter_column

    def wrap_expr_column(self, fieldname):
        """Return a default expr for the field

        :param fieldname: name of the field
        """
        attr_name = anyblok_column_prefix + fieldname

        def expr_column(model_self):
            return getattr(model_self, attr_name)

        return expr_column

    def expire_related_attribute(self, model_self, action_todos):
        for action_todo in action_todos:
            if len(action_todo) == 1:
                obj = model_self
                attrs = [action_todo[0]]
            else:
                obj = getattr(model_self, action_todo[0])
                attrs = [action_todo[1]]
                if obj is None:
                    continue

            if obj in model_self.registry.session:
                if obj._sa_instance_state.persistent:
                    model_self.registry.expire(obj, attrs)

    def setter_format_value(self, value):
        return value

    def wrap_setter_column(self, fieldname):
        attr_name = anyblok_column_prefix + fieldname

        def setter_column(model_self, value):
            action_todos = set()
            if fieldname in model_self.loaded_columns:
                action_todos = model_self.registry.expire_attributes.get(
                    model_self.__registry_name__, {}).get(fieldname, set())

            self.expire_related_attribute(model_self, action_todos)
            value = self.setter_format_value(value)
            res = setattr(model_self, attr_name, value)
            self.expire_related_attribute(model_self, action_todos)
            return res

        return setter_column

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """ Return the instance of the real field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known of the model
        :rtype: instance of Field
        """
        self.format_label(fieldname)
        return self

    def format_label(self, fieldname):
        """ Return the label for this field

        :param fieldname: if no label filled, the fieldname will be capitalized
            and returned
        :rtype: the label for this field
        """
        if not self.label:
            label = fieldname.replace('_', ' ')
            self.label = label.capitalize()

    def native_type(self):
        """ Return the native SqlAlchemy type

        :exception: FieldException
        """
        raise FieldException("No native type for this field")

    def must_be_declared_as_attr(self):
        """ Return False, it is the default value """
        return False

    def must_be_duplicate_before_added(self):
        """ Return False, it is the default value """
        return False

    def autodoc_get_properties(self):
        res = {'Type': self.__class__}
        res['Context'] = self.context
        res['Label'] = self.label
        res.update(self.kwargs)
        return res

    autodoc_omit_property_values = set((
        ('Label', None),
        ('Context', None),
        ))

    def autodoc_format_dict(self, key, value, level=0):
        bullets = ['*', '+', '•', '‣']
        bullet = bullets[level]
        padding = '  ' * level
        key = key.strip()
        if isinstance(value, dict):
            res = padding + '%c ``%s``:\n\n' % (bullet, key)
            res += '\n'.join(
                [self.autodoc_format_dict(x, y, level=level + 1)
                 for x, y in value.items()])
            res += '\n'
            return res
        elif isinstance(value, (list, tuple)):
            res = padding + '%c ``%s``:\n\n' % (bullet, key)
            next_bullet = bullets[level + 1]
            res += '\n'.join(padding + '  %c ``%r``' % (next_bullet, x)
                             for x in value)
            res += '\n'
            return res
        else:
            if isinstance(value, type):
                rst_val = ':class:`%s.%s`' % (value.__module__,
                                              value.__name__)
            elif isinstance(value, ModelRepr):
                rst_val = value.model_name
            else:
                rst_val = '``%r``' % value
            return padding + '%c ``%s`` - %s' % (bullet, key, rst_val)

    def autodoc_do_omit(self, k, v):
        """Maybe convert, then check in :attr:`autodoc_omit_property_values`

        Mutable types aren't hashable, and usually, if not empty, it makes
        sense to display them. Therefore, we replace them by None if
        empty to decide and let through otherwise.

        Hence, to exclude empty Context from autodoc, is done by putting
        ``('Context', None)`` in :attr:`autodoc_omit_property_values`
        """
        if isinstance(v, list) or isinstance(v, dict) or isinstance(v, set):
            if v:
                return True
            v = None
        return (k, v) in self.autodoc_omit_property_values

    def autodoc(self):
        return '\n'.join(self.autodoc_format_dict(x, y)
                         for x, y in self.autodoc_get_properties().items()
                         if not self.autodoc_do_omit(x, y))


class Function(Field):
    """ Function Field

    ::

        from anyblok.declarations import Declarations
        from anyblok.field import Function


        @Declarations.register(Declarations.Model)
        class Test:
            x = Function(fget='fget', fset='fset', fdel='fdel', fexp='fexpr')

        ..warning::

            fexp must be a classmethod

    """

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """Return the property of the field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """

        def wrap(method):
            m = self.kwargs.get(method)
            if m is None:
                return None

            def function_method(model_self, *args, **kwargs):
                if method == 'fget':
                    cls = registry.get(model_self.__registry_name__)
                    if model_self is cls:
                        return hasattr(model_self, m)

                return getattr(model_self, m)(*args, **kwargs)

            return function_method

        fget = wrap('fget')
        fset = wrap('fset')
        fdel = wrap('fdel')
        fexpr = wrap('fexpr')

        self.format_label(fieldname)
        properties['loaded_fields'][fieldname] = self.label
        return hybrid_property(fget, fset, fdel=fdel, expr=fexpr)


def format_struc(entry, keys):
    key = keys[0]
    if len(keys) == 1:
        if key not in entry:
            entry[key] = None
    else:
        if key not in entry:
            entry[key] = {}

        format_struc(entry[key], keys[1:])


class JsonRelated(Field):
    """ Json Related Field

    ::

        from anyblok.declarations import Declarations
        from anyblok.field import JsonRelated
        from anyblok.column import Json


        @Declarations.register(Declarations.Model)
        class Test:
            properties = Json()
            x = JsonRelated(json_column='properties', keys=['x'])

    """
    def __init__(self, *args, **kwargs):
        self.json_column = kwargs.pop('json_column', None)
        if self.json_column is None:
            raise FieldException(
                "json_column is a required attribute for "
                "JsonRelated"
            )
        self.keys = kwargs.pop('keys', None)
        if self.keys is None:
            raise FieldException(
                "keys is a required attribute for JsonRelated"
            )
        self.get_adapter = kwargs.pop('get_adapter', None)
        self.set_adapter = kwargs.pop('set_adapter', None)
        super(JsonRelated, self).__init__(*args, **kwargs)

    def get_fget(self):
        def fget(model_self):
            json_column = getattr(model_self, self.json_column)
            if json_column is None:
                json_column = {}

            format_struc(json_column, self.keys)
            entry = json_column
            for key in self.keys:
                entry = entry[key]

            if entry and self.get_adapter:
                get_adapter = self.get_adapter
                if isinstance(get_adapter, str):
                    get_adapter = getattr(model_self, get_adapter)

                entry = get_adapter(entry)

            return entry

        return fget

    def get_fset(self):
        def fset(model_self, value):
            json_column = getattr(model_self, self.json_column)
            if json_column is None:
                json_column = {}

            format_struc(json_column, self.keys)
            entry = json_column
            for key in self.keys[:-1]:
                entry = entry[key]

            if value and self.set_adapter:
                set_adapter = self.set_adapter
                if isinstance(set_adapter, str):
                    set_adapter = getattr(model_self, set_adapter)

                value = set_adapter(value)

            entry[self.keys[-1]] = value
            setattr(model_self, self.json_column, json_column)

        return fset

    def get_fdel(self):
        def fdel(model_self):
            json_column = getattr(model_self, self.json_column)
            if json_column is None:
                json_column = {}

            format_struc(json_column, self.keys)
            entry = json_column
            for key in self.keys[:-1]:
                entry = entry[key]

            entry[self.keys[-1]] = None
            setattr(model_self, self.json_column, json_column)

        return fdel

    def get_fexpr(self):
        def fexpr(model_self):
            entry = getattr(model_self, self.json_column)
            for key in self.keys:
                entry = entry[key]

            return entry

        return fexpr

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """Return the property of the field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: properties known to the model
        """
        self.format_label(fieldname)
        properties['loaded_fields'][fieldname] = self.label
        return hybrid_property(
            self.get_fget(),
            self.get_fset(),
            fdel=self.get_fdel(),
            expr=self.get_fexpr()
        )

    def autodoc_get_properties(self):
        res = super(JsonRelated, self).autodoc_get_properties()
        res.update({
            'json_column': self.json_column,
            'keys': ' -> '.join(self.keys),
        })
        return res



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import XMLImporterException
from lxml import etree
from os import urandom


class TestImportCSV(BlokTestCase):

    def create_importer(self, file_to_import=None, **kwargs):
        XML = self.registry.IO.Importer.XML
        if 'model' not in kwargs:
            kwargs['model'] = 'Model.IO.Importer'

        if file_to_import is None:
            file_to_import = urandom(100000)
        return XML.insert(file_to_import=file_to_import, **kwargs)

    def create_XML_importer(self, blokname=None, **kwargs):
        XML = self.registry.IO.Importer.XML
        return XML(self.create_importer(**kwargs), blokname=blokname)

    def test_commit_if_error_found(self):
        importer = self.create_XML_importer()
        importer.error_found.append('Mock error')
        self.assertEqual(importer.commit(), False)

    def test_commit(self):
        importer = self.create_XML_importer()
        self.assertEqual(importer.commit(), True)

    def test_raise_now(self):
        importer = self.create_XML_importer()
        msg = 'test'
        with self.assertRaises(XMLImporterException):
            importer._raise(msg, on_error='raise')

        self.assertIn(msg, importer.error_found)

    def test_raise_at_the_end(self):
        importer = self.create_XML_importer()
        msg = 'test'
        importer._raise(msg)
        self.assertIn(msg, importer.error_found)

    def test_raise_ignore(self):
        importer = self.create_XML_importer()
        msg = 'test'
        importer._raise(msg, on_error="ignore")
        self.assertIn(msg, importer.error_found)

    def test_find_entry(self):
        importer = self.create_XML_importer()
        self.assertEqual(importer.find_entry(), None)

    def test_find_entry_params(self):
        importer = self.create_XML_importer()
        model = 'Model.System.Model'
        param = 'test_param'
        importer.params[(model, param)] = importer
        self.assertEqual(importer.find_entry(model=model, param=param),
                         importer)

    def test_find_entry_external_id(self):
        importer = self.create_XML_importer()
        param = 'test_param'
        self.registry.IO.Mapping.set(param, importer.importer)
        self.assertEqual(
            importer.find_entry(model=importer.importer.__registry_name__,
                                external_id=param), importer.importer)

    def test_check_entry_before_import(self):
        importer = self.create_XML_importer()
        record = etree.Element('record')
        self.assertEqual(importer.check_entry_before_import(record, importer),
                         (False, importer))
        self.assertEqual(importer.error_found, [])

    def test_check_entry_before_import_if_exist_pass(self):
        importer = self.create_XML_importer()
        record = etree.Element('record')
        self.assertEqual(importer.check_entry_before_import(record,
                                                            importer,
                                                            if_exist='pass'),
                         (True, importer))
        self.assertEqual(importer.error_found, [])

    def test_check_entry_before_import_if_exist_raise(self):
        importer = self.create_XML_importer()
        record = etree.Element('record')
        self.assertEqual(importer.check_entry_before_import(record,
                                                            importer,
                                                            if_exist='raise'),
                         (True, None))
        self.assertEqual(len(importer.error_found), 1)

    def test_check_entry_before_import_if_does_not_exist_pass(self):
        importer = self.create_XML_importer()
        record = etree.Element('record')
        self.assertEqual(importer.check_entry_before_import(
            record, None, if_does_not_exist='pass'), (True, None))
        self.assertEqual(len(importer.error_found), 0)

    def test_check_entry_before_import_if_does_not_exist_raise(self):
        importer = self.create_XML_importer()
        record = etree.Element('record')
        self.assertEqual(importer.check_entry_before_import(
            record, None, if_does_not_exist='raise'), (True, None))
        self.assertEqual(len(importer.error_found), 1)

    def test_import_entry_if_exist_continue(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        self.assertEqual(importer.import_entry(importer, None, model=model,
                                               if_exist='continue'),
                         importer)
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_import_entry_if_exist_create(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        values = {
            'model': model,
            'mode': model + '.XML',
            'file_to_import': urandom(100000),
        }
        self.assertNotEqual(importer.import_entry(importer, values,
                                                  model=model,
                                                  if_exist='create'),
                            importer)
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_import_entry_if_exist_update(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        values = {
            'model': model,
            'mode': model + '.XML',
            'file_to_import': urandom(100000),
        }
        self.assertEqual(importer.import_entry(importer.importer, values,
                                               model=model),
                         importer.importer)
        self.assertEqual(len(importer.created_entries), 0)
        self.assertEqual(len(importer.updated_entries), 1)
        self.assertEqual(len(importer.error_found), 0)

    def test_import_entry_if_does_not_exist_create(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        values = {
            'model': model,
            'mode': model + '.XML',
            'file_to_import': urandom(100000),
        }
        importer.import_entry(None, values, model=model, if_exist='create')
        self.assertEqual(len(importer.created_entries), 1)
        self.assertEqual(len(importer.updated_entries), 0)
        self.assertEqual(len(importer.error_found), 0)

    def test_import_entry_external_id(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.import_entry(importer.importer, None, model=model,
                              if_exist='continue', external_id='test_import')
        self.assertEqual(self.registry.IO.Mapping.get(model, 'test_import'),
                         importer.importer)

    def test_import_entry_external_id_with_blokname(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.import_entry(importer.importer, None, model=model,
                              if_exist='continue', external_id='test_import')
        mapping = self.registry.IO.Mapping.get(model, 'test_import')
        mapping.blokname = 'Test'

    def test_import_entry_mapper(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.import_entry(importer, None, model=model,
                              if_exist='continue', param='test_import')
        self.assertEqual(importer.params[(model, 'test_import')], importer)

    def test_import_entry_mapper_exist(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.params[(model, 'test_import')] = importer
        importer.import_entry(importer, None, model=model,
                              if_exist='continue', param='test_import')
        self.assertEqual(importer.params[(model, 'test_import')], importer)

    def test_import_entry_mapper_exist_and_not_the_same(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.params[(model, 'test_import')] = importer.importer
        importer.import_entry(importer, None, model=model,
                              if_exist='continue', param='test_import')
        self.assertEqual(len(importer.error_found), 1)

    def test_import_multi_value(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        record1 = etree.Element('field')
        record1.set('model', model)
        field = etree.SubElement(record1, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record1, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'
        record2 = etree.Element('field')
        record2.set('model', model)
        field = etree.SubElement(record2, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record2, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        res = importer.import_multi_values([record1, record2], model)
        self.assertEqual(len(res), 2)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(len(importer.created_entries), 2)

    def test_import_value(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        record = etree.Element('field')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        res = importer.import_value(record, 'Integer', model)
        self.assertTrue(isinstance(res, int))
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(len(importer.created_entries), 1)

    def test_import_value_relation_ship(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        record = etree.Element('field')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        res = importer.import_value(record, 'Many2One', model)
        self.assertTrue(isinstance(res.id, int))
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(len(importer.created_entries), 1)

    def test_import_field(self):
        importer = self.create_XML_importer()
        field = etree.Element('field')
        field.set('name', 'model')
        field.text = 'Model.IO.Importer'
        self.assertEqual(importer.import_field(field, 'String'), field.text)

    def test_import_field_external_id(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        self.registry.IO.Mapping.set('test_import', importer.importer)
        field = etree.Element('field')
        field.set('name', 'model')
        field.set('external_id', 'test_import')
        self.assertEqual(importer.import_field(field, 'String', model=model),
                         importer.importer.id)
        self.assertEqual(len(importer.error_found), 0)

    def test_import_field_external_id_without_model(self):
        importer = self.create_XML_importer()
        self.registry.IO.Mapping.set('test_import', importer.importer)
        field = etree.Element('field')
        field.set('name', 'model')
        field.set('external_id', 'test_import')
        self.assertEqual(importer.import_field(field, 'String'), None)
        self.assertEqual(len(importer.error_found), 1)

    def test_import_field_params(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.params[(model, 'test_import')] = importer
        field = etree.Element('field')
        field.set('name', 'model')
        field.set('param', 'test_import')
        self.assertEqual(importer.import_field(field, 'String', model=model),
                         importer)
        self.assertEqual(len(importer.error_found), 0)

    def test_import_field_inexisting_params(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.params[(model, 'test_import')] = model
        field = etree.Element('field')
        field.set('name', 'model')
        field.set('param', 'test_import')
        field.text = 'Model.IO.Importer'
        self.assertEqual(importer.import_field(field, 'String', model=model),
                         field.text)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(importer.params[(model, 'test_import')], field.text)

    def test_import_field_params_without_model(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Importer'
        importer.params[(model, 'test_import')] = importer
        field = etree.Element('field')
        field.set('name', 'model')
        field.set('param', 'test_import')
        self.assertEqual(importer.import_field(field, 'String'), None)
        self.assertEqual(len(importer.error_found), 1)

    def test_import_record_without_model(self):
        importer = self.create_XML_importer()
        record = etree.Element('record')
        self.assertEqual(importer.import_record(record), None)
        self.assertEqual(len(importer.error_found), 1)

    def test_import_record_with_wrong_node(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        record = etree.Element('record')
        record.set('model', model)
        etree.SubElement(record, 'badnode')
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        importer.import_record(record)
        self.assertEqual(len(importer.error_found), 1)

    def test_import_record_field_without_name(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        record = etree.Element('record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        importer.import_record(record)
        self.assertEqual(len(importer.error_found), 1)

    def test_import_record(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        record = etree.Element('record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        importer.import_record(record)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(len(importer.created_entries), 1)

    def test_import_records(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        records = etree.Element('records')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        importer.import_records(records)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(len(importer.created_entries), 1)

    def test_import_records_with_bad_node(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        records = etree.Element('records')
        etree.SubElement(records, 'badnode')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'

        importer.import_records(records)
        self.assertEqual(len(importer.error_found), 1)
        self.assertEqual(len(importer.created_entries), 1)

    def test_import_records_with_commit(self):
        importer = self.create_XML_importer()
        model = 'Model.IO.Exporter'
        records = etree.Element('records')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'
        etree.SubElement(records, 'commit')

        importer.import_records(records)
        self.assertEqual(len(importer.error_found), 0)
        self.assertEqual(len(importer.created_entries), 1)

    def test_run_raise(self):
        model = 'Model.IO.Exporter'
        records = etree.Element('records')
        records.set('on_error', 'raise')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'
        file_to_import = etree.tostring(records)
        importer = self.create_XML_importer(file_to_import=file_to_import)
        with self.assertRaises(XMLImporterException):
            importer.run()

        self.assertEqual(len(importer.error_found), 1)

    def test_run_raise_ignore(self):
        model = 'Model.IO.Exporter'
        records = etree.Element('records')
        records.set('on_error', 'ignore')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'
        file_to_import = etree.tostring(records)
        importer = self.create_XML_importer(file_to_import=file_to_import)
        res = importer.run()
        self.assertEqual(len(res['error_found']), 1)

    def test_run(self):
        model = 'Model.IO.Exporter'
        records = etree.Element('records')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'
        file_to_import = etree.tostring(records)
        importer = self.create_XML_importer(file_to_import=file_to_import)
        res = importer.run()
        self.assertEqual(len(res['error_found']), 0)
        self.assertEqual(len(res['created_entries']), 1)
        self.assertEqual(len(res['updated_entries']), 0)

    def test_run_bad_root_name(self):
        model = 'Model.IO.Exporter'
        records = etree.Element('badrootname')
        records.set('on_error', 'ignore')
        record = etree.SubElement(records, 'record')
        record.set('model', model)
        field = etree.SubElement(record, 'field')
        field.set('name', 'model')
        field.text = model
        field = etree.SubElement(record, 'field')
        field.set('name', 'mode')
        field.text = model + '.XML'
        file_to_import = etree.tostring(records)
        importer = self.create_XML_importer(file_to_import=file_to_import)
        res = importer.run()
        self.assertEqual(len(res['error_found']), 1)
        self.assertEqual(len(res['created_entries']), 0)
        self.assertEqual(len(res['updated_entries']), 0)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import threading
from inspect import ismethod


class EnvironmentException(AttributeError):
    """ Exception for the Environment """


class EnvironmentManager:
    """ Manage the Environment for an application """

    environment = None

    @classmethod
    def define_environment_cls(cls, Environment):
        """ Define the class used for the environment

        :param Environment: class of environment
        :exception: EnvironmentException
        """

        def check_classmethod(method, acceptNone=False):
            if not hasattr(Environment, method):
                raise EnvironmentException("No %r found" % method)

            m = getattr(Environment, method)

            if m is None:
                if acceptNone:
                    return

            if ismethod(m):
                return

            raise EnvironmentException("%r must be a class method" % m)

        check_classmethod('scoped_function_for_session', acceptNone=True)
        check_classmethod('setter')
        check_classmethod('getter')

        cls.environment = Environment

    @classmethod
    def set(cls, key, value):
        """ Save the value of the key in the environment

        :param key: the key of the value to save
        :param value: the value to save
        :exception: EnvironmentException
        """
        if cls.environment is None:
            raise EnvironmentException("No environments defined")

        cls.environment.setter(key, value)

    @classmethod
    def get(cls, key, default=None):
        """ Load the value of the key in the environment

        :param key: the key of the value to load
        :param default: return this value if not value loaded for the key
        :rtype: the value of the key
        :exception: EnvironmentException
        """
        if cls.environment is None:
            raise EnvironmentException("No environments defined")

        return cls.environment.getter(key, default)

    @classmethod
    def scoped_function_for_session(cls):
        """ Save the value of the key in the environment """
        return cls.environment.scoped_function_for_session


class ThreadEnvironment:
    """ Use the thread, to get the environment """

    scoped_function_for_session = None
    """ No scoped function here because for none value sqlalchemy already uses
    a thread to save the session """

    values = {}

    @classmethod
    def setter(cls, key, value):
        """ Save the value of the key in the environment

        :param key: the key of the value to save
        :param value: the value to save
        """
        if str(threading.current_thread()) not in cls.values:
            cls.values[str(threading.current_thread())] = {}

        cls.values[str(threading.current_thread())][key] = value

    @classmethod
    def getter(cls, key, default):
        """ Get the value of the key in the environment

        :param key: the key of the value to retrieve
        :param default: return this value if no value loaded for the key
        :rtype: the value of the key
        """
        if str(threading.current_thread()) not in cls.values:
            return default

        return cls.values[str(threading.current_thread())].get(key, default)


EnvironmentManager.define_environment_cls(ThreadEnvironment)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok import Declarations


register = Declarations.register
IO = Declarations.Model.IO


@register(IO)
class Exporter:

    @classmethod
    def get_mode_choices(cls):
        res = super(Exporter, cls).get_mode_choices()
        res.update({'Model.IO.Exporter.XML': 'XML'})
        return res


@register(IO.Exporter)
class XML:

    def __init__(self, exporter):
        self.exporter = exporter

    @classmethod
    def insert(cls, delimiter=None, quotechar=None, fields=None, **kwargs):
        kwargs['mode'] = cls.__registry_name__
        if 'model' in kwargs:
            if not isinstance(kwargs['model'], str):
                kwargs['model'] = kwargs['model'].__registry_name__

        return cls.registry.IO.Exporter.insert(**kwargs)

    def run(self, entries):
        raise NotImplemented



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import DBTestCase


class TestCoreQuery(DBTestCase):

    def test_inherit(self):

        def inherit():

            from anyblok import Declarations
            Core = Declarations.Core

            @Declarations.register(Core)
            class Query:

                def foo(self):
                    return True

        registry = self.init_registry(inherit)
        self.assertEqual(registry.System.Blok.query().foo(), True)



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.tests.testcase import BlokTestCase
from ..exceptions import CSVExporterException
from csv import DictReader


class TestExportCSV(BlokTestCase):

    def create_exporter(self, Model, **kwargs):
        Exporter = self.registry.IO.Exporter.CSV
        return Exporter.insert(model=Model, **kwargs)

    def test_create_exporter_by_registry_name(self):
        exporter = self.create_exporter('Model.IO.Exporter')
        self.assertEqual(exporter.model, 'Model.IO.Exporter')

    def test_create_exporter_by_model(self):
        exporter = self.create_exporter(self.registry.IO.Exporter)
        self.assertEqual(exporter.model, 'Model.IO.Exporter')

    def test_create_exporter_with_field(self):
        fields = [{'name': 'model'}]
        exporter = self.create_exporter(self.registry.IO.Exporter,
                                        fields=fields)
        self.assertEqual(exporter.model, 'Model.IO.Exporter')
        self.assertEqual(len(exporter.fields_to_export), 1)
        self.assertEqual(exporter.fields_to_export[0].name, 'model')

    def test_create_exporter_with_two_fields(self):
        fields = [{'name': 'model'}, {'name': 'csv_delimiter'}]
        exporter = self.create_exporter(self.registry.IO.Exporter,
                                        fields=fields)
        self.assertEqual(exporter.model, 'Model.IO.Exporter')
        self.assertEqual(len(exporter.fields_to_export), 2)

    def test_get_header_from_fields_any(self):
        fields = [{'name': 'id'}]
        exporter = self.create_exporter(self.registry.IO.Exporter,
                                        fields=fields)
        res = exporter.get_model(exporter.mode)(exporter).get_header()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], 'id')

    def test_get_header_from_fields_external_id(self):
        fields = [{'name': 'id', 'mode': 'external_id'}]
        exporter = self.create_exporter(self.registry.IO.Exporter,
                                        fields=fields)
        res = exporter.get_model(exporter.mode)(exporter).get_header()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], 'id/EXTERNAL_ID')

    def test_format_field_with_mapping_any(self):
        Exporter = self.registry.IO.Exporter
        fields = [{'name': 'id'}]
        exporter = self.create_exporter(Exporter, fields=fields)
        res = exporter.fields_to_export[0].value2str(exporter, exporter)
        self.assertEqual(res, str(exporter.id))

    def test_format_field_with_mapping_external_id(self):
        Exporter = self.registry.IO.Exporter
        fields = [{'name': 'id', 'mode': 'external_id'}]
        exporter = self.create_exporter(Exporter, fields=fields)
        key = Exporter.get_external_id(exporter.model)
        key = key.split('_')
        key = '_'.join([key[0], str(int(key[1]) + 1)])
        res = exporter.fields_to_export[0].value2str(exporter, exporter)
        self.assertEqual(res, key)

    def test_format_field_with_forbidden_mapping_external_id(self):
        Exporter = self.registry.IO.Exporter
        fields = [{'name': 'id.other', 'mode': 'external_id'}]
        exporter = self.create_exporter(Exporter, fields=fields)
        with self.assertRaises(CSVExporterException):
            exporter.fields_to_export[0].value2str(exporter, exporter)

    def test_format_browsed_field_with_mapping_any(self):
        Exporter = self.registry.IO.Exporter
        fields = [{'name': 'model.is_sql_model'}]
        exporter = self.create_exporter(Exporter, fields=fields)
        res = exporter.fields_to_export[0].value2str(exporter, exporter)
        self.assertEqual(res, '1')

    def test_format_browsed_field_pks_without_mapping_key(self):
        Exporter = self.registry.IO.Exporter
        fields = [{'name': 'model.name'}]
        exporter = self.create_exporter(Exporter, fields=fields)
        res = exporter.fields_to_export[0].value2str(exporter, exporter)
        self.assertEqual(res, 'Model.IO.Exporter')

    def test_format_browsed_field_with_mapping_key(self):
        Exporter = self.registry.IO.Exporter
        fields = [{'name': 'model.name', 'mode': 'external_id'}]
        exporter = self.create_exporter(Exporter, fields=fields)
        model = self.registry.System.Model.from_primary_keys(
            name=Exporter.__registry_name__)
        key = self.registry.IO.Exporter.get_key_mapping(model)
        res = exporter.fields_to_export[0].value2str(exporter, exporter)
        self.assertEqual(res, key)

    def test_export_anyblok_core(self):
        Blok = self.registry.System.Blok
        fields = [{'name': 'name', 'mode': 'external_id'},
                  {'name': 'state'}]
        exporter = self.create_exporter(Blok, fields=fields)
        blok = Blok.from_primary_keys(name="anyblok-core")
        key = self.registry.IO.Exporter.get_key_mapping(blok)
        csvfile = exporter.run([blok])
        reader = DictReader(csvfile, delimiter=exporter.csv_delimiter,
                            quotechar=exporter.csv_quotechar)
        rows = [x for x in reader]
        csvfile.close()
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0]['name/EXTERNAL_ID'], key)
        self.assertEqual(rows[0]['state'], 'installed')

    def test_export_all_bloks(self):
        Blok = self.registry.System.Blok
        fields = [{'name': 'name'}, {'name': 'state'}, {'name': 'version'},
                  {'name': 'installed_version'}, {'name': 'order'}]
        exporter = self.create_exporter(Blok, fields=fields)
        bloks = Blok.query().all()
        exporter.run(bloks)
